@@ -117,7 +117,16 @@ def cacheable(fn):
 
 
 def base_defines_strategy(force_reusable):
+    """Returns a decorator for strategy functions.
+
+    If force_reusable is True, the generated values are assumed to be
+    reusable, i.e. immutable and safe to cache, across multiple test
+    invocations.
+
+    """
     def decorator(strategy_definition):
+        """A decorator that registers the function as a strategy and makes it
+        lazily evaluated."""
         from hypothesis.searchstrategy.lazy import LazyStrategy
         _strategies.add(strategy_definition.__name__)
 
@@ -192,7 +201,7 @@ def just(value):
     return JustStrategy(value)
 
 
-@defines_strategy
+@defines_strategy_with_reusable_values
 def none():
     """Return a strategy which only generates None.
 
@@ -241,8 +250,8 @@ def integers(min_value=None, max_value=None):
     If min_value is not None then all values will be >= min_value. If
     max_value is not None then all values will be <= max_value
 
-    Examples from this strategy will shrink towards being positive (e.g. 1000
-    is considered simpler than -1) and then towards zero.
+    Examples from this strategy will shrink towards zero, and negative values
+    will also shrink towards positive (i.e. -n may be replaced by +n).
 
     """
 
@@ -529,22 +538,24 @@ def lists(
             def unique_by(x):
                 return x
 
+    if min_size is None:
+        min_size = 0
+
+    if max_size is None:
+        max_size = float('inf')
+
+    if average_size is None:
+        average_size = max(
+            _AVERAGE_LIST_LENGTH,
+            min_size * 2
+        )
+        if max_size < float('inf'):
+            average_size = min(average_size, 0.5 * (min_size + max_size))
+
+    from hypothesis.searchstrategy.collections import ListStrategy, \
+        UniqueListStrategy
     if unique_by is not None:
-        from hypothesis.searchstrategy.collections import UniqueListStrategy
-        min_size = min_size or 0
-        max_size = max_size or float(u'inf')
-        if average_size is None:
-            if max_size < float(u'inf'):
-                if max_size <= 5:
-                    average_size = min_size + 0.75 * (max_size - min_size)
-                else:
-                    average_size = (max_size + min_size) / 2
-            else:
-                average_size = max(
-                    _AVERAGE_LIST_LENGTH,
-                    min_size * 2
-                )
-        result = UniqueListStrategy(
+        return UniqueListStrategy(
             elements=elements,
             average_size=average_size,
             max_size=max_size,
@@ -552,20 +563,10 @@ def lists(
             key=unique_by
         )
     else:
-        from hypothesis.searchstrategy.collections import ListStrategy
-        if min_size is None:
-            min_size = 0
-        if average_size is None:
-            if max_size is None:
-                average_size = _AVERAGE_LIST_LENGTH
-            else:
-                average_size = (min_size + max_size) * 0.5
-
-        result = ListStrategy(
+        return ListStrategy(
             (elements,), average_length=average_size,
             min_size=min_size, max_size=max_size,
         )
-    return result
 
 
 @cacheable
@@ -936,23 +937,41 @@ def random_module():
 
 @cacheable
 @defines_strategy
-def builds(target, *args, **kwargs):
+def builds(*callable_and_args, **kwargs):
     """Generates values by drawing from ``args`` and ``kwargs`` and passing
-    them to ``target`` in the appropriate argument position.
+    them to the callable (provided as the first positional argument) in the
+    appropriate argument position.
 
     e.g. ``builds(target, integers(), flag=booleans())`` would draw an
     integer ``i`` and a boolean ``b`` and call ``target(i, flag=b)``.
 
-    If ``target`` has type annotations, they will be used to infer a strategy
+    If the callable has type annotations, they will be used to infer a strategy
     for required arguments that were not passed to builds.  You can also tell
     builds to infer a strategy for an optional argument by passing the special
     value :const:`hypothesis.infer` as a keyword argument to
-    builds, instead of a strategy for that argument to ``target``.
+    builds, instead of a strategy for that argument to the callable.
 
     Examples from this strategy shrink by shrinking the argument values to
-    the target.
+    the callable.
 
     """
+    if callable_and_args:
+        target, args = callable_and_args[0], callable_and_args[1:]
+        if not callable(target):
+            raise InvalidArgument(
+                'The first positional argument to builds() must be a callable '
+                'target to construct.')
+    elif 'target' in kwargs and callable(kwargs['target']):
+        args = []
+        note_deprecation(
+            'Specifying the target as a keyword argument to builds() is '
+            'deprecated. Provide it as the first positional argument instead.')
+        target = kwargs.pop('target')
+    else:
+        raise InvalidArgument(
+            'builds() must be passed a callable as the first positional '
+            'argument, but no positional arguments were given.')
+
     if infer in args:
         # Avoid an implementation nightmare juggling tuples and worse things
         raise InvalidArgument('infer was passed as a positional argument to '
@@ -972,30 +991,16 @@ def builds(target, *args, **kwargs):
     )
 
 
-def delay_error(func):
-    """A decorator to make exceptions lazy but success immediate.
-
-    We want from_type to resolve to a strategy immediately if possible,
-    for a useful repr and interactive use, but delay errors until a
-    value would be drawn to localise them to a particular test.
-
-    """
+def _defer_from_type(func):
+    """Decorator to make from_type lazy to support recursive definitions."""
     @proxies(func)
     def inner(*args, **kwargs):
-        try:
-            return func(*args, **kwargs)
-        except Exception as e:
-            error = e
-
-            def lazy_error():
-                raise error
-
-            return builds(lazy_error)
+        return deferred(lambda: func(*args, **kwargs))
     return inner
 
 
 @cacheable
-@delay_error
+@_defer_from_type
 def from_type(thing):
     """Looks up the appropriate search strategy for the given type.
 
@@ -1024,12 +1029,12 @@ def from_type(thing):
 
     """
     from hypothesis.searchstrategy import types
-    if not isinstance(thing, type):
-        try:
+    try:
+        import typing
+        if not isinstance(thing, type):
             # At runtime, `typing.NewType` returns an identity function rather
             # than an actual type, but we can check that for a possible match
             # and then read the magic attribute to unwrap it.
-            import typing
             if all([
                 hasattr(thing, '__supertype__'), hasattr(typing, 'NewType'),
                 isfunction(thing), getattr(thing, '__module__', 0) == 'typing'
@@ -1040,8 +1045,16 @@ def from_type(thing):
             if getattr(thing, '__origin__', None) is typing.Union:
                 args = sorted(thing.__args__, key=types.type_sorting_key)
                 return one_of([from_type(t) for t in args])
-        except ImportError:  # pragma: no cover
-            pass
+        # We can't resolve forward references, and under Python 3.5 (only)
+        # a forward reference is an instance of type.  Hence, explicit check:
+        elif type(thing) == typing._ForwardRef:  # pragma: no cover
+            raise ResolutionFailed(
+                'thing=%s cannot be resolved.  Upgrading to python>=3.6 may '
+                'fix this problem via improvements to the typing module.'
+                % (thing,))
+    except ImportError:  # pragma: no cover
+        pass
+    if not isinstance(thing, type):
         raise InvalidArgument('thing=%s must be a type' % (thing,))
     # Now that we know `thing` is a type, the first step is to check for an
     # explicitly registered strategy.  This is the best (and hopefully most
@@ -1518,6 +1531,7 @@ def composite(f):
     """
 
     from hypothesis.internal.reflection import define_function_signature
+    from hypothesis.searchstrategy.strategies import calc_label
     argspec = getfullargspec(f)
 
     if (
@@ -1536,6 +1550,8 @@ def composite(f):
               if k in (argspec.args + argspec.kwonlyargs + ['return'])}
     new_argspec = argspec._replace(args=argspec.args[1:], annotations=annots)
 
+    label = calc_label(f)
+
     @defines_strategy
     @define_function_signature(f.__name__, f.__doc__, new_argspec)
     def accept(*args, **kwargs):
@@ -1549,6 +1565,10 @@ def composite(f):
                     return data.draw(strategy)
 
                 return f(draw, *args, **kwargs)
+
+            @property
+            def label(self):
+                return label
         return CompositeStrategy()
     accept.__module__ = f.__module__
     return accept
@@ -1801,16 +1821,16 @@ def deferred(definition):
     >>> x.example()
     (((False, (True, True)), (False, True)), (True, True))
     >>> x.example()
-    (True, True)
+    True
 
     Mutual recursion also works fine:
 
     >>> a = st.deferred(lambda: st.booleans() | b)
     >>> b = st.deferred(lambda: st.tuples(a, a))
     >>> a.example()
-    (True, (True, False))
+    True
     >>> b.example()
-    (False, True)
+    (False, (False, ((False, True), False)))
 
     Examples from this strategy shrink as they normally would from the strategy
     returned by the definition.
